"""Perform steady-state JV simulations"""

######### Package Imports #########################################################################

import os, uuid, sys
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from scipy import constants
# import pySIMsalabim
## Import pySIMsalabim, if not successful, add the parent directory to the system path
try :
    import pySIMsalabim as sim
except ImportError:
    # Add the parent directory to the system path
    sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '../..')))
    import pySIMsalabim as sim
from pySIMsalabim.utils import general as utils_gen
from pySIMsalabim.utils.parallel_sim import *
from pySIMsalabim.utils.utils import update_cmd_pars

######### Functions #################################################################################

def run_SS_JV(simss_device_parameters, session_path, JV_file_name = 'JV.dat', varFile = 'none', G_fracs = [], parallel = False, max_jobs = max(1,os.cpu_count()-1), run_mode = True, **kwargs):
    """

    Parameters
    ----------
    simss_device_parameters : string
        Name of the device parameters file.
    session_path : string
        Path to the session folder where the simulation will run.
    JV_file_name : string
        Name of the JV file.
    parallel : bool, optional
        Run the simulations in parallel, by default False
    max_jobs : int, optional
        Maximum number of parallel jobs, by default max(1,os.cpu_count()-1)
    cmd_pars : _type_, optional
        _description_, by default None
    UUID : str, optional
        _description_, by default ''
    run_mode : bool, optional
        indicate whether the script is in 'web' mode (True) or standalone mode (False). Used to control the console output, by default True
    **kwargs : dict
        Additional arguments to be passed to the function.

    Returns
    -------
    int
        Exitcode of the simulation.
    str
        Message from the simulation.

    """
    verbose = kwargs.get('verbose', False) # Check if the user wants to show verbose output in the console
    UUID = kwargs.get('UUID', '') # Check if the user wants to add a UUID to the JV file name
    cmd_pars = kwargs.get('cmd_pars', None) # Check if the user wants to add additional command line parameters to the 
    force_multithreading = kwargs.get('force_multithreading', False) # Check if the user wants to force multithreading instead of using GNU parallel
    cmd_pars = kwargs.get('cmd_pars', None) # Check if the user wants to add additional command line parameters
    # Check if the user wants to force the use of thread safe mode, necessary for Windows with parallel simulations
    if os.name == 'nt':  
        threadsafe = kwargs.get('threadsafe', True) # Check if the user wants to force the use of threads instead of processes
    else:
        threadsafe = kwargs.get('threadsafe', False) # Check if the user wants to force the use of threads instead of processes

    turnoff_autoTidy = kwargs.get('turnoff_autoTidy', None) # Check if the user wants to turn off the autoTidy function in SIMsalabim
    if turnoff_autoTidy is None: 
        if not threadsafe:
            turnoff_autoTidy = True
        else:
            turnoff_autoTidy = False

    # Update the JV file name with the UUID
    if UUID != '':
        dum_str = f'_{UUID}'
    else:
        dum_str = ''
    
    # Define the command to be executed
    if G_fracs is None:
        # Update the filenames with the UUID
        JV_file_name = os.path.join(session_path,JV_file_name)
        logFile = os.path.join(session_path,'log.txt')
        scParsFile = os.path.join(session_path,'scPars.txt')
        if UUID != '':
            JV_file_name_base, JV_file_name_ext = os.path.splitext(JV_file_name)
            JV_file_name = JV_file_name_base + dum_str + JV_file_name_ext
            logFile = os.path.join(session_path,'log'+dum_str+'.txt')
            scParsFile = os.path.join(session_path,'scPars'+dum_str+'.txt')
        if varFile != 'none':
            var_file_base, var_file_ext = os.path.splitext(varFile)
            varFile = var_file_base + dum_str + var_file_ext
            varFile = os.path.join(session_path,varFile)


        
        # Specify the arguments to be attached to the cmd
        SS_JV_args = [{'par':'dev_par_file','val':simss_device_parameters},
                        {'par':'JVFile','val':JV_file_name},
                        {'par':'logFile','val':logFile},
                        {'par':'scParsFile','val':scParsFile}
                        ]
        
        if turnoff_autoTidy:
            SS_JV_args.append({'par':'autoTidy','val':'0'})

        if varFile != 'none':
            SS_JV_args.append({'par':'varFile','val':varFile})
        else:
            SS_JV_args.append({'par':'outputRatio','val':'0'})

        # Update the cmd_pars with the SS_JV_args
        if cmd_pars is not None:
            SS_JV_args = update_cmd_pars(SS_JV_args, cmd_pars)

        if threadsafe:
            # Run the simulation in thread safe mode
            result, message = utils_gen.run_simulation_filesafe('simss', SS_JV_args, session_path, run_mode = run_mode,verbose=verbose)
        else:
            result, message = utils_gen.run_simulation('simss',SS_JV_args,session_path,run_mode = run_mode,verbose=verbose)

        return result, message

    else:
        # Update the filenames with the UUID
        JV_file_name = os.path.join(session_path,JV_file_name)
        JV_file_name_base, JV_file_name_ext = os.path.splitext(JV_file_name)
        if varFile != 'none':
            var_file_base, var_file_ext = os.path.splitext(varFile)
            varFile = var_file_base + dum_str + var_file_ext
            varFile = os.path.join(session_path,varFile)

        # SS_JV_args = [{'par':'dev_par_file','val':simss_device_parameters}]
        SS_JV_args_list = []
        for G_frac in G_fracs:
            dum_args = [{'par':'dev_par_file','val':simss_device_parameters},
            {'par':'G_frac','val':str(G_frac)},
                                    {'par':'JVFile','val':JV_file_name_base + f'_Gfrac_{G_frac}' + dum_str + JV_file_name_ext},
                                    {'par':'logFile','val':os.path.join(session_path,'log'+f'_Gfrac_{G_frac}'+dum_str+'.txt')},
                                    {'par':'scParsFile','val':os.path.join(session_path,'scPars'+f'_Gfrac_{G_frac}'+dum_str+'.txt')},
                                    {'par':'varFile','val':os.path.join(session_path,var_file_base + f'_Gfrac_{G_frac}' + dum_str +var_file_ext)} if varFile != 'none' else {'par':'varFile','val':'none'},
                                    ]
            if varFile == 'none':
                dum_args.append({'par':'outputRatio','val':'0'})

            if turnoff_autoTidy:
                dum_args.append({'par':'autoTidy','val':'0'})
                
            if cmd_pars is not None:
                dum_args = update_cmd_pars(dum_args, cmd_pars)

            SS_JV_args_list.append(dum_args)                             
                                       
        if parallel and len(G_fracs) > 1:
            results = run_simulation_parallel('simss', SS_JV_args_list, session_path, max_jobs, force_multithreading=force_multithreading,verbose=verbose)
            msg_list = ['' for i in range(len(results))]
        else:
            results, msg_list = [], []
            for dum_args in SS_JV_args_list:

                if threadsafe:
                    result, message = utils_gen.run_simulation_filesafe('simss', dum_args, session_path, run_mode,verbose=verbose)
                else:
                    result, message = utils_gen.run_simulation('simss', dum_args, session_path, run_mode,verbose=verbose)
                
                results.append(result)
                msg_list.append(message)
        
        # check if results is a list of CompletedProcess objects
        if isinstance(results, list) :
            if len(results) > 0 and not isinstance(results[0], subprocess.CompletedProcess):
                pass
            else:    
                if len(results) > 0 and isinstance(results[0], tuple) and all(isinstance(res[0], subprocess.CompletedProcess) for res in results):
                    # Extract the return codes from the CompletedProcess objects
                    results = [res[0].returncode for res in results]
        
        # Check if all simulations were successful
        if all([res == 0 for res in results]):
            if verbose and not run_mode:
                print('All JV simulations completed successfully\n')
                # for mess in msg_list:
                #     print(mess)
            return 0, 'All JV simulations completed successfully'
        elif all([(res == 0 or res == 95) for res in results]):
            if verbose and not run_mode:
                print('All JV simulations completed successfully, but some had some points that did not converge\n')
                for mess in msg_list:
                    print(mess)
            return 0, 'All JV simulations completed successfully, but some had some points that did not converge'
        else:
            if verbose and not run_mode:
                print('Some JV simulations failed\n')
                for i, res in enumerate(results):
                    print(f'Simulation {i+1} failed with return code {res}')

            # get all results that are not 0
            failed_results = [res for i, res in enumerate(results) if res != 0 and res != 95]
            # If there is only one failed result, return it with the error message
            # get only unique values in failed_results
            failed_results = list(set(failed_results))
            if len(failed_results) == 1:
                return failed_results[0], utils_gen.error_message(failed_results[0])
            else:
                print(failed_results)
                # if len(failed_results) > 1:
                print(f"Multiple different errors occurred during the parallel simulations: {set([val for val in failed_results if val not in [0, 95, 3]])}. Returning error code 666.")
                return 666, utils_gen.error_message(666)
                # else:

            

## Running the function as a standalone script
if __name__ == "__main__":
    ## Manual Steady State input parameters. These are overwritten by command line arguments if provided
    # G_fracs = [0.9, 1.0]
    G_fracs = None

    # Define folder and file paths
    session_path = os.path.join('../../','SIMsalabim','SimSS')

    simss_device_parameters = 'simulation_setup.txt'

    JV_name = 'JV.dat'
    Var_name = 'Var.dat'

    # UUID = str(uuid.uuid4()) # Add a UUID to the simulation
    UUID = ''

    # Not user input
    parallel = False
    run_mode = False # If False, show verbose output in console

    ############## Command line arguments  ##############
    ## Notes
    ## - The command line arguments are optional and can be provided in any order
    ## - Each command line argument must be provided in the format -par_name value
    ## - Possible arguments include all SIMsalabim parameters and Steady State JV specific parameters as listed before
    ## - Special arguments
    ##   - G_fracs : string
    ##     - The G fractions to simulate, separated by a comma
    ##   - sp : string
    ##     - The session path, i.e. the working directory for the simulation
    ##   - simsetup : string
    ##     - The name of the simss simulation setup parameters file
    ##   - UUID : string
    ##     - An UUID to add to the simulation (output)

    cmd_pars_dict = {}
    cmd_pars = []

    # Check if any arguments are provided.
    if len(sys.argv) >= 2:
        # Each arguments should be in the format -par val, i.e. in pairs of two. Skip the first argument as this is the script name
        if not len(sys.argv[1:]) % 2 == 0:
            print('Error in command line parameters. Please provide arguments in the format -par_name value -par_name value ...')
            sys.exit(1)
        else:
            input_list = sys.argv[1:]
            # Loop over the input list and put pairs into a dictionary with the first argument as key and the second argument as value
            for i in range(0,len(input_list),2):
                # Check if the key already exists, if not, add it to the cmd_pars_dict
                if str(input_list[i][1:]) in cmd_pars_dict:
                    print(f'Duplicate parameter found in the command line parameters: {str(input_list[i][1:])}')
                    sys.exit(1)
                else:
                    cmd_pars_dict[str(input_list[i][1:])] = str(input_list[i+1])

    # Check and process specific keys/arguments that are not native SIMsalabim arguments
    # Handle the session_path/sp argument separately, as the other parameters depend on this
    if 'sp' in cmd_pars_dict:
        session_path = cmd_pars_dict['sp']
        # remove from cmd_pars
        cmd_pars_dict.pop('sp')

    # Define mappings for keys and variables
    key_action_map = {
        'simsetup': lambda val: {'zimt_device_parameters': val},
        'G_fracs': lambda val: {'G_fracs': val.split(',')}, # IMPORTANT: if multiple Gfracs are provided, they must be separated by a comma!!
        'JV_name': lambda val: {'JV_name': val},
        'Var_name': lambda val: {'Var_name': val},
        'UUID': lambda val: {'UUID': val},
    }

    for key in list(cmd_pars_dict.keys()):  # Use list to avoid modifying the dictionary while iterating
        if key in key_action_map:
            # Apply the corresponding action
            result = key_action_map[key](cmd_pars_dict[key])
            globals().update(result)  # Dynamically update global variables
            cmd_pars_dict.pop(key)

    # Handle remaining keys in `cmd_pars_dict` and add them to the cmd_pars list
    cmd_pars.extend({'par': key, 'val': value} for key, value in cmd_pars_dict.items())

    # Run the function
    results, message = run_SS_JV(simss_device_parameters, session_path, JV_name, Var_name, G_fracs, parallel = parallel, run_mode = run_mode, cmd_pars=cmd_pars, UUID=UUID)

    # Print the results if simulation failed
    if run_mode:
        if G_fracs == None:
            print(message)
        else:
            if results != 0:
                for i, msg in enumerate(message):
                    print(f'G_frac: {G_fracs[i]} - {msg}')



###################################################
############### Useful function ###################
###################################################
# by Vincent M. Le Corre
# Package import
import subprocess,shutil,os,glob,sys,warnings
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from scipy import stats,optimize,constants
from lmfit import Model
# Don't show warnings
warnings.filterwarnings("ignore")
## Physics constants
q = constants.value(u'elementary charge')
eps_0 = constants.value(u'electric constant')
kb = constants.value(u'Boltzmann constant in eV/K')
k = constants.value(u'Boltzmann constant')
#############################################################
################ SCLC analysis functions ####################
#############################################################

def MottGurney(V,mu,eps_r,Vbi,L):
    """Mott-Gurney equation typically used to fit single carrier devices JVs to get the mobility.

    J = (9/8)*eps_0*eps_r*mu*((V-Vbi)**2)/(L**3)

    Refs:
    Mott N F and Gurney R W 1940 Electronic Processes in Ionic Crystals (Oxford: Oxford University Press)
    Nice paper to read on MG ==> J. Phys.: Condens. Matter 30 (2018) 105901

    Parameters
    ----------
    V : 1-D sequence of floats
        Array containing the voltages (V)
    mu : float
        Mobility (m^2/(Vs))
    eps_r : float
        Relative dielectric constant
    Vbi : float
        Built-in voltage (V)
    L : float
        Thickness (m)

    Returns
    -------
    1-D sequence of floats
        Array containing the current-density (A m^-2)
    """    
    return (9/8)*eps_0*eps_r*mu*((V-Vbi)**2)/(L**3)


def fit_MottGurney(V,J,mu,eps_r,Vbi,L,var2fit=['mu']):
    """Fit the Mott-Gurney equation to the data

    Parameters
    ----------
    V : 1-D sequence of floats
        Array containing the voltages (V)
    J : 1-D sequence of floats
        Array containing the current-density (A m^-2)
    mu : float
        Mobility (m^2/(Vs))
    eps_r : float
        Relative dielectric constant
    Vbi : float
        Built-in voltage (V)
    L : float
        Thickness (m)
    var2fit : list, optional
        Chose variable to fit (can be multiple), by default ['mu']

    Returns
    -------
    list
        list of the fitted values in the same order than var2fit
    """    
    fixed_var = ['V','mu','eps_r','Vbi','L']
    for i in var2fit:
        if i in fixed_var:
            fixed_var.remove(i)
    
    Mod_MottGurney = Model(MottGurney, independent_vars=fixed_var)
    result = Mod_MottGurney.fit(J, V=V,mu=mu,eps_r=eps_r,Vbi=Vbi,L=L)
   
    best_fit = []
    for i in var2fit:
        best_fit.append(result.params[i].value)
    return best_fit

def deriv(x,y):
    """Get the derivative of input data dy/dx

    Parameters
    ----------
    x : 1-D sequence of floats
        x-axis data
    y : 1-D sequence of floats
        y-axis data

    Returns
    -------
    1-D sequence of floats
        dy/dx
    """    
    dy = np.diff(y)/np.diff(x)
    dum = (y[-1] - y[-2])/(x[-1] - x[-2]) 
    dy = np.append(dy,[dum])
    # dum = (y[1] - y[0])/(x[1] - x[0]) 
    # dy = np.append([dum],dy)
    return list(dy)

def log_slope(x,y):
    """Get the slope of the logarithmic data

    Parameters
    ----------
    x : 1-D sequence of floats
        x-axis data
    y : 1-D sequence of floats
        y-axis data

    Returns
    -------
    1-D sequence of floats
        slope of the logarithmic data
    """ 
    # Convert data into np.arrays
    x = np.asarray(x)
    y = np.asarray(y)   
    # remove V = 0V before getting log
    if 0 in x:
        idx_0 = x.index(0)
        x = x.pop(idx_0)
        y = y.pop(idx_0)
    log_xf = np.log(x)
    log_yf = np.log(y)
    # Take derivative of data
    slopef = deriv(log_xf,log_yf)
    return np.asarray(slopef)

def calc_Vsat(L,Nc,phi,eps_r,T):
    """ Calculate Vsat for SCLC measurement 
    as defined in equation (15) in https://doi.org/10.1021/acsenergylett.0c02599

    Vsat = (8/9) * ((q*Nc*L**2)/(eps_0*eps_r))*exp(-q*phi/k*T)

    Parameters
    ----------
    L : float
        Thickness (m)
    Nc : float
        Effective density of states (m^-3)
    phi : float
        Injection barrier (V)
    eps_r : float
        Relative dielectric constant
    T : float
        Temperature (K)

    Returns
    -------
    float
        returns Vsat as defined in 
    """    
    Vsat = (8/9) * ((q*Nc*L**2)/(eps_0*eps_r))*np.exp(-(q*phi)/(k*T))
    return Vsat

def calc_Vtfl(traps,L,eps_r):
    """Calculate VTFL for SCLC measurement 
    as defined in equation (2) in ACS Energy Lett. 2021, 6, 3, 1087–1094
    https://doi.org/10.1021/acsenergylett.0c02599

    Vnet = q * n_traps * L**2 /( 2 * eps_0 * eps_r) 

    Parameters
    ----------
    traps : float
        Trap density (m^-3)
    L : float
        Thickness (m)
    eps_r : float
        Relative dielectric constant

    Returns
    -------
    float
        returns vnet as defined in 
    """    
    vtfl = q * (traps) * L**2 /( 2 * eps_0 * eps_r)
    return vtfl

def calc_trap_charge(Vnet,L,eps_r):
    """Calculate the trap charges for SCLC measurement
    as defined in equation (2) in ACS Energy Lett. 2021, 6, 3, 1087–1094
    https://doi.org/10.1021/acsenergylett.0c02599

    Vnet = q * n_trap * L**2 /( 2 * eps_0 * eps_r) 

    Note that if there are not ions or dopant n_net = traps

    Parameters
    ----------
    Vtfl : float
        Vtfl see ref
    L : float
        Thickness (m)
    eps_r : float
        Relative dielectric constant

    Returns
    -------
    float
        trapped charges in the bulk
    """    
    n_trap = Vtfl * ( 2 * eps_0 * eps_r)/ (q * L**2)
    return n_trap # m^-3

def calc_Vnet_with_ions(ions,traps,L,eps_r):
    """Calculate Vnet for SCLC measurement in the presence of Ions
    as defined in equation (4) in ACS Energy Lett. 2021, 6, 3, 1087–1094
    https://doi.org/10.1021/acsenergylett.0c02599

    Vnet = q * n_net * L**2 /( 2 * eps_0 * eps_r) = q * (traps-ions) * L**2 /( 2 * eps_0 * eps_r)

    Parameters
    ----------
    ions : float
        Ion density (m^-3)
    traps : float
        Trap density (m^-3)
    L : float
        Thickness (m)
    eps_r : float
        Relative dielectric constant

    Returns
    -------
    float
        returns vnet as defined in 
    """    
    vnet = q * (traps-ions) * L**2 /( 2 * eps_0 * eps_r)
    return vnet

def calc_net_charge(Vnet,L,eps_r):
    """Calculate the net charge for SCLC measurement
    as defined in equation (4) in ACS Energy Lett. 2021, 6, 3, 1087–1094
    https://doi.org/10.1021/acsenergylett.0c02599

    Vnet = q * n_net * L**2 /( 2 * eps_0 * eps_r) 

    Note that if there are not ions or dopant n_net = traps

    Parameters
    ----------
    Vnet : float
        Vnet see ref
    L : float
        Thickness (m)
    eps_r : float
        Relative dielectric constant

    Returns
    -------
    float
        net-charge in the bulk
    """    
    net_charge= Vnet * ( 2 * eps_0 * eps_r)/ (q * L**2)
    return net_charge # m^-3

def calc_nt_min(L,eps_r,T):
    """Calculate the minimum appearant trap density for SCLC measurement
    as defined in equation (3) in ACS Energy Lett. 2021, 6, 3, 1087–1094
    https://doi.org/10.1021/acsenergylett.0c02599

    Ntmin = 4*np.Pi()**2*((k*T)/(q**2))*((eps_0*eps_r)/(L**2))

    Parameters
    ----------
    L : float
        Thickness (m)
    eps_r : float
        Relative dielectric constant
    T : float
        Temperature (K)

    Returns
    -------
    float
        minimum appearant trap density
    """    
    Ntmin = 4*np.pi**2*((k*T)/(q**2))*((eps_0*eps_r)/(L**2))
    return Ntmin # m^-3


def SCLC_get_data_plot(volt,curr):
    """Get the data for the SCLC plot

    Parameters
    ----------
    volt : 1-D sequence of floats
        Voltage data
    curr : 1-D sequence of floats
        Current data

    Returns
    -------
    V_slopef : 1-D sequence of floats
        Voltage data after filtering
    J_slopef : 1-D sequence of floats
        Current data after filtering
    slopesf : 1-D sequence of floats
        logarithmic slopes of the JV after filtering
    get_tangentf : bool
        True if tangents are calculated i.e. if max(slopef) > 2, False if not
    idx_maxf : int
        index of the max slope of the JV after filtering
    max_slopesf : float
        max slope of the JV after filtering
    tang_val_V1f : 1-D sequence of floats
        tangent values of the JV after filtering for the first tangent point. Typically crossing between the ohmic and Trap filled limited regime.
    tang_val_V2f : 1-D sequence of floats
        tangent values of the JV after filtering for the second tangent point (here the inflexion point)
    tang_val_V3f : 1-D sequence of floats
        tangent values of the JV after filtering for the third tangent point. Typically crossing between the Trap filled limited regime and the Mott-Gurney regime.
    V1f : float
        Voltage value of the first crossing point
    J1f : float
        current value of the first crossing point
    V2f : float
        Voltage value of the second crossing point
    J2f : float
        current value of the second crossing point
    Vinf : float
        Voltage value of the inflexion point
    Jinf : float
        current value of the inflexion point
        
    """    
    
    volt = np.asarray(volt)
    curr = np.asarray(curr)

    # remove volt = 0V before getting log
    if 0 in volt:
        idx_0 = np.where(volt == 0)[0][0]
        volt = np.delete(volt,idx_0)
        curr = np.delete(curr,idx_0)
    
    # if we get a 0 or negative value in volt or curr raise an error
    if 0 in volt or 0 in curr:
        raise ValueError('We got a 0 value in the voltage or current data, please check your input data')
    if any(i < 0 for i in volt) or any(i < 0 for i in curr):
        raise ValueError('We got a negative value in the voltage or current data, please check your input data')

    # remove first offset_noise points in case it is noisy
    # offset_noise = 0
    # volt = volt[offset_noise:]
    # curr = curr[offset_noise:]

    slopesf = log_slope(volt,curr)
    V_slopef = np.asarray(volt)
    J_slopef = np.asarray(curr)

    # Get max slope if max slopes > 2.2
    idx_maxf = np.argmax(slopesf) 
    max_slopesf = slopesf[idx_maxf]

    if max_slopesf>2.0:
        get_tangentf = 1
    else: get_tangentf = 0


    # Calculate tangent and crossing point
    tang_val_V1f,tang_val_V2f,tang_val_V3f,V1f,J1f,V2f,J2f,Vinf,Jinf = [],[],[],[],[],[],[],[],[]
    if get_tangentf == 1:
            tang_val_V1f = np.exp(np.log(J_slopef[0])+slopesf[0]*(np.log(V_slopef)-np.log(V_slopef[0])))
            tang_val_V2f = np.exp(np.log(J_slopef[idx_maxf])+max_slopesf*(np.log(V_slopef)-np.log(V_slopef[idx_maxf])))
            tang_val_V3f = np.exp(np.log(J_slopef[-1])+slopesf[-1]*(np.log(V_slopef)-np.log(V_slopef[-1])))
            V1f = np.exp((1/(slopesf[0]-max_slopesf)*np.log((J_slopef[idx_maxf]*(V_slopef[0]**slopesf[0]))/(J_slopef[0]*(V_slopef[idx_maxf]**max_slopesf)))))
            J1f = np.exp(np.log(J_slopef[0])+slopesf[0]*(np.log(V1f)-np.log(V_slopef[0])))
            V2f = np.exp((1/(slopesf[-1]-max_slopesf)*np.log((J_slopef[idx_maxf]*(V_slopef[-1]**slopesf[-1]))/(J_slopef[-1]*(V_slopef[idx_maxf]**max_slopesf)))))
            J2f = np.exp(np.log(J_slopef[-1])+slopesf[-1]*(np.log(V2f)-np.log(V_slopef[-1])))
            Vinf = V_slopef[idx_maxf]
            Jinf = J_slopef[idx_maxf]

    # print(calc_net_charge(V1,thick))
    # print(calc_net_charge(V2,thick))
   
    return V_slopef,J_slopef,slopesf,get_tangentf,idx_maxf,max_slopesf,tang_val_V1f,tang_val_V2f,tang_val_V3f,V1f,J1f,V2f,J2f,Vinf,Jinf

def Make_SCLC_plot(ax,data_JV,x='Vext',y=['Jext'],show_tangent=[1,2,3],xlimits=[],ylimits=[],plot_type=0,labels='',colors='b',line_type = ['-'],mark='',legend=True,plot_jvexp=False,data_JVexp=pd.DataFrame(),save_yes=False,pic_save_name='JV.jpg'):
    """ Make typical plot for SCLC measurement analysis for SIMsalabim  
    
    Parameters
    ----------
    num_fig : int
        number of the fig to plot JV

    data_JV : DataFrame
        Panda DataFrame containing JV_file

    x : str, optional
        xaxis data  (default = 'Vext'), by default 'Vext'

    y : list of str, optional
        yaxis data can be multiple like ['Jext','Jbimo']  (default = ['Jext']), by default ['Jext']
    
    show_tangent : list of int, optional
        show tangent line at V1,Vinf,V2  (default = [1,2,3]), by default [1,2,3]

    xlimits : list, optional
        x axis limits if = [] it lets python chose limits, by default []

    ylimits : list, optional
        y axis limits if = [] it lets python chose limits, by default []

    plot_type : int, optional
        type of plot 1 = logx, 2 = logy, 3 = loglog else linlin (default = linlin), by default 0

    labels : str, optional
        label of the JV, by default ''

    colors : str, optional
        color for the JV line, by default 'b'

    line_type : list, optional
        type of line for simulated data plot
        size line_type need to be = size(y), by default ['-']

    mark : str, optional
        type of Marker for the JV, by default ''

    legend : bool, optional
        Display legend or not, by default True

    plot_jvexp : bool, optional
        plot an experimental JV or not, by default False

    data_JVexp : [type], optional
        Panda DataFrame containing experimental JV_file with 'V' the voltage and 'J' the current, by default pd.DataFrame()

    save_yes : bool, optional
        If True, save JV as an image with the  file name defined by "pic_save_name", by default False

    pic_save_name : str, optional
        name of the file where the figure is saved, by default 'JV.jpg'
    
    Returns
    -------
    Vslopef : list
        list of of the Voltage values after filtering
    Jslopef : list
        list of of the current values after filtering
    slopesf : list
        list of logarithmic slopes of the JV after filtering
    get_tangentf : bool
        True if tangents are calculated i.e. if max(slopef) > 2, False if not
    idx_maxf : int
        index of the max slope of the JV after filtering
    max_slopesf : float
        max slope of the JV after filtering
    tang_val_V1f : list
        list of the tangent values of the JV after filtering for the first tangent point. Typically crossing between the ohmic and Trap filled limited regime.
    tang_val_V2f : list
        list of the tangent values of the JV after filtering for the second tangent point (here the inflexion point)
    tang_val_V3f : list
        list of the tangent values of the JV after filtering for the third tangent point. Typically crossing between the Trap filled limited regime and the Mott-Gurney regime.
    V1f : float
        Voltage value of the first crossing point
    J1f : float
        current value of the first crossing point
    V2f : float
        Voltage value of the second crossing point
    J2f : float
        current value of the second crossing point
    Vinf : float
        Voltage value of the inflexion point
    Jinf : float
        current value of the inflexion point

    """    

    # Filter data for V < 0 and J < 0
    print('In Make_SCLC_plot')
    print('Filtering data for V < 0 and J < 0')
    print('If you want to plot JV with V < 0, you need to change modify your input first')
    data_JV = data_JV[data_JV.Vext > 0]
    data_JV = data_JV[data_JV.Jext > 0]

    # Get tangent and crossing point data
    if show_tangent != []:
        V_slopef,J_slopef,slopesf,get_tangentf,idx_maxf,max_slopesf,tang_val_V1f,tang_val_V2f,tang_val_V3f,V1f,J1f,V2f,J2f,Vinf,Jinf = SCLC_get_data_plot(np.asarray(data_JV['Vext']),np.asarray(data_JV['Jext']))
    else :
        get_tangentf = False                
    if len(y) > len(line_type):
        print('Invalid line_type list, we meed len(y) == len(line_type)')
        print('We will use default line type instead')
        line_type = []
        for counter, value in enumerate(y):
            line_type.append('-')

    # plt.figure(num_fig)
    # ax_JVs_plot = plt.axes()
    ax_JVs_plot = ax
    for i,line in zip(y,line_type):
        if plot_type == 1:
            ax_JVs_plot.semilogx(data_JV['Vext'],data_JV[i]/10,color=colors,label=labels,linestyle=line,marker=mark,markeredgecolor=colors,markersize=15,markerfacecolor='None',markeredgewidth = 3)
            if plot_jvexp:
                ax_JVs_plot.semilogx(data_JVexp['V'],data_JVexp['J']/10,'o',markeredgecolor=colors,markersize=15,markerfacecolor='None',markeredgewidth = 3)
            
        
        elif plot_type == 2:
            ax_JVs_plot.semilogy(data_JV['Vext'],data_JV[i]/10,color=colors,label=labels,linestyle=line,marker=mark,markeredgecolor=colors,markersize=15,markerfacecolor='None',markeredgewidth = 3)   
            if plot_jvexp:
                ax_JVs_plot.semilogy(data_JVexp['V'],data_JVexp['J']/10,'o',markeredgecolor=colors,markersize=15,markerfacecolor='None',markeredgewidth = 3)
            
        elif plot_type == 3:
            ax_JVs_plot.loglog(data_JV['Vext'],data_JV[i]/10,color=colors,label=labels,linestyle=line,marker=mark,markeredgecolor=colors,markersize=15,markerfacecolor='None',markeredgewidth = 3)
            if plot_jvexp:
                ax_JVs_plot.loglog(data_JVexp['V'],data_JVexp['J']/10,'o',markeredgecolor=colors,markersize=15,markerfacecolor='None',markeredgewidth = 3)
            
            if get_tangentf and show_tangent != []:
                print('Plotting tangents and crossing points')
                if 1 in show_tangent:
                    ax_JVs_plot.loglog(V_slopef,tang_val_V1f/10,color=colors,linestyle=line)
                if 2 in show_tangent:
                    ax_JVs_plot.loglog(V_slopef,tang_val_V2f/10,color=colors,linestyle=line)
                if 3 in show_tangent:
                    ax_JVs_plot.loglog(V_slopef,tang_val_V3f/10,color=colors,linestyle=line)
                if 1 in show_tangent and 2 in show_tangent:
                    ax_JVs_plot.loglog(V1f,J1f/10,'rs',markersize=15)
                if 2 in show_tangent and 3 in show_tangent:
                    ax_JVs_plot.loglog(V2f,J2f/10,'m^',markersize=15)
                if 2 in show_tangent:
                    ax_JVs_plot.loglog(Vinf,Jinf/10,'bo',markersize=15)
            
        else:
            ax_JVs_plot.plot(data_JV['Vext'],data_JV[i]/10,color=colors,label=labels,linestyle=line,marker=mark,markeredgecolor=colors,markersize=15,markerfacecolor='None',markeredgewidth = 3)
            if plot_jvexp:
                ax_JVs_plot.plot(data_JVexp['V'],data_JVexp['J']/10,'o',markeredgecolor=colors,markersize=15,markerfacecolor='None',markeredgewidth = 3)
        
    
    # legend
    if legend == True:
        plt.legend(loc='best',ncol=2,frameon=False)#,fontsize = 30)
    if xlimits != []:
        plt.xlim(xlimits)
    if ylimits != []:
        plt.ylim(ylimits)
    plt.grid(b=True,which='both')
    plt.xlabel('Applied Voltage [V]')
    plt.ylabel('Current Density [mA cm$^{-2}$]')
    plt.tight_layout()
    if save_yes:
        plt.savefig(pic_save_name,dpi=100,transparent=True)
    
    if show_tangent != []:
        return ax,V_slopef,J_slopef,slopesf,get_tangentf,idx_maxf,max_slopesf,tang_val_V1f,tang_val_V2f,tang_val_V3f,V1f,J1f,V2f,J2f,Vinf,Jinf





import os,math
import pandas as pd

def read_tj_file(session_path, tj_file_name='tj.dat'):
    """ Read relevant parameters for impedance of the tj file

    Parameters
    ----------
    session_path : string
        Path of the simulation folder for this session
    data_tj : dataFrame
        Pandas dataFrame containing the tj output file from ZimT

    Returns
    -------
    DataFrame
        Pandas dataFrame containing the time, voltage, current density and numerical error in the current density of the tj_file
    """

    data = pd.read_csv(os.path.join(session_path,tj_file_name), sep=r'\s+')

    return data

def get_integral_bounds(data, f_min=1e-2, f_max=1e6, f_steps=20):
    """ Determine integral bounds in the time domain, used to compute the conductance and capacitance

    Parameters
    ----------
    data : dataFrame
        Pandas dataFrame containing the time, voltage, current density and numerical error in the current density of the tj_file
    f_min : float
        Minimum frequency
    f_max : float
        Maximum frequency
    f_steps : float
        Frequency steps

    Returns
    -------
    list
        List of array indices that will be used in the plotting
    """

    # Total number of time points
    numTimePoints = len(data['t'])

    # Check which time index corresponds to 1/fmax. We call this istart:
    istart = -1
    for i in range(numTimePoints):
        if math.isclose(data['t'][i], 1/f_max, rel_tol = 2/f_steps): #note: don't use == to compare 2 floating points!
            istart = i

    # Starting time point could not be found.
    if istart == -1:
        msg = 'Could not find a time that corresponds to the highest frequency.'
        return -1, msg
    
    # print('Found istart: ', istart)

    # ifin: last index we should plot, corresponds to time = 1/f_min:
    ifin = numTimePoints - 1

    # isToPlot starts with istart:
    isToPlot = [istart]

    PlotRatio = max(1, round( (ifin-istart)/(math.log10(f_max/f_min) * f_steps)))

    # Incorrect plot ratio
    if PlotRatio < 1:
        msg = 'PlotRatio smaller than 1. It should at least be 1'
        return -1, msg

    # Then add the other indices:
    for i in range(istart+1, ifin-1):
        if (i-istart) % PlotRatio == 0: # note: % is python's modulo operator.
            isToPlot.append(i) # add the index to our array

    # Also include the last index:
    isToPlot.append(ifin)

    # Integral bounds have been determined, return the array with indices and a success message
    msg = 'Success'
    return isToPlot, msg

def update_cmd_pars(main_pars, cmd_pars):
    """Merges main parameters with command line parameters.

    Parameters
    ----------
    main_pars : list of dict
        A list containing dictionaries of the main parameters of the application. These parameters
        serve as the default values.
    cmd_pars : list of dict
        A list containing dictionaries of parameters passed through the command line interface.
        These parameters have higher precedence and override the main parameters in case
        of a conflict within any dictionary.

    Returns
    -------
    list of dict
        A list of dictionaries containing the combined set of parameters, with command line parameters
        overriding main parameters in case of conflicts within any dictionary.

    Raises
    ------
    ValueError
        If duplicate parameters are found in the command line parameters.
    """

    # check for duplicate in cmd_pars
    new_pars_names = []
    for line in cmd_pars:
        new_pars_names.append(line['par'])
    # if duplicates are found raise an error
    if len(new_pars_names) != len(set(new_pars_names)):
        raise ValueError('Duplicate parameters found in the command line parameters')

    # Update the main parameters with the command line parameters
    for par in cmd_pars:
        found = False
        for main in main_pars:
            if par['par'] == main['par']:
                main['val'] = str(par['val']) # convert to string
                break
        if not found:
            main_pars.append(par)          
    
    return main_pars




"""Global settings for plots"""
######### Package Imports #########################################################################

import matplotlib.pyplot as plt

######### Function Definitions ####################################################################

# List of default colors
color = ['#1f77b4', '#ff7f0e', '#2ca02c', '#d62728', '#9467bd', '#8c564b', '#e377c2', '#7f7f7f', '#bcbd22', '#17becf']

# Plot parameters and sizes
plt.rc('font', size = 16)
plt.rc('axes', linewidth=1, titlesize=14)     # fontsize of the axes title
plt.rc('axes', labelsize=14)    # fontsize of the x and y labels
plt.rc('xtick', top=True, bottom=True, direction='in')
plt.rc('ytick', left=True, right=True, direction='in') 
plt.rc('figure', titlesize=5)
plt.rc('legend', fontsize=14)
plt.rc('lines', linewidth=2)


"""Compile program with fpc"""
# Description:
# ------------
# Compile program using fpc

######### Package Imports ###################################################################

import os,subprocess

######### Function Definitions ##############################################################

def fpc_prog(prog_name,path2prog,show_term_output=True,force_fpc=True,verbose=True):
    """Compile program using fpc

    Parameters
    ----------
    prog_name : str
        Program name (e.g. 'SimSS', 'zimt', 'SIMsalabim)
    path2prog : str
        String of the absolute path to the program
    show_term_output : bool, optional
        show terminal output from the compilation, by default True
    force_fpc : bool, optional  
        force recompile with fpc even if compiled program already exists, by default True
    verbose : bool, optional
        print output of the compilation, by default True
    """   

    is_windows = (os.name == 'nt')          # Check if we are on Windows
    path2prog = str(path2prog)                  # Convert to string
    # Check if the program is already compiled
    if (os.path.isfile(os.path.join(path2prog,prog_name+'.exe')) and is_windows) or (os.path.isfile(os.path.join(path2prog,prog_name)) and not is_windows):
        if force_fpc:
            if show_term_output == True:
                output_direct = None
            else:
                output_direct = subprocess.DEVNULL
            try:
                subprocess.check_call(['fpc', prog_name.lower()+'.pas'], encoding='utf8', stdout=output_direct, cwd=path2prog, shell=is_windows)
            except subprocess.CalledProcessError:
                print('Error compiling '+prog_name+' in '+path2prog)
                raise ChildProcessError
            if verbose:
                print('\n'+prog_name+' already existed but was recompiled'+'\n')
        else:
            if verbose:  
                print('\n'+prog_name+' already compiled')
    
    else: # Compile the program
        if show_term_output == True:
            output_direct = None
        else:
            output_direct = subprocess.DEVNULL
        try:
            subprocess.check_call(['fpc', prog_name.lower()+'.pas'], encoding='utf8', stdout=output_direct, cwd=path2prog, shell=is_windows)
            if verbose:
                print('\n'+prog_name+' was not compiled so we did it!'+'\n')
        except subprocess.CalledProcessError:
            print('Error compiling '+prog_name+' in '+path2prog)
            raise ChildProcessError    

    


if __name__ == '__main__':

    print('I am not doing anything, please import me in your script'+'\n')
    
#     System = platform.system()                  # Operating system
#     is_windows = (System == 'Windows')          # Check if we are on Windows
#     if is_windows:
#         fpc_prog('SimSS', 'c:/Users/lecor/Desktop/GitHub/SIMsalabim/SimSS')
#         fpc_prog('zimt', 'c:/Users/lecor/Desktop/GitHub/SIMsalabim/ZimT')
#     else:
#         fpc_prog('SimSS', '/mnt/c/Users/lecor/Desktop/GitHub/SIMsalabim/SimSS')
#         fpc_prog('zimt', '/mnt/c/Users/lecor/Desktop/GitHub/SIMsalabim/ZimT')


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pySIMsalabim", # Replace with your own username
    version="v1.03",
    author="Vincent M. Le Corre, Sander Heester, L. Jan Anton Koster",
    author_email="",
    description="",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license='GPLv3',
    url="https://github.com/kostergroup/pySIMsalabim",
    download_url="https://github.com/kostergroup/pySIMsalabim/v1.03.tar.gz",
    packages=setuptools.find_packages(),
    readme = "README.md",
    keywords=['Drift-diffusion', 'semiconductor', 'solar cells'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Intended Audience :: Science/Research",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    ],
    python_requires='>=3.12',
    install_requires = [
        'numpy>=1.2, <=2.0',
        'pandas>=1.4',
        'matplotlib>=3.5',
        'jupyterlab>=3.4',
        'seaborn>=0.11',
        'scipy>=1.0',
        'gitpython>=3.1',
        'openpyxl>=3.0',
    ],
    extras_require = {
        'dev': [
            'pytest',
            'twine',
        ],
    },
    include_package_data=True,
        
)


"""Functions for plotting"""
######### Package Imports #########################################################################

import matplotlib
import matplotlib.pyplot as plt
from matplotlib.collections import LineCollection
import numpy as np
from pySIMsalabim.plots import plot_def

######### Function Definitions ####################################################################

def plot_result(data, pars, selected, x_key, xlabel, ylabel, xscale, yscale, title, ax, plot_funcs, x_error = [], y_error=[], legend=True, error_fmt='-'):
    """Make a plot for a (sub)set of parameters from a DataFrame. Note: errorbars only work with single x,y functions


    Parameters
    ----------
    data : DataFrame
        Data to plot
    pars : dict
        Dict with all potential parameters to plot. Keys represent the names in the dataFrama, values are the corresponding labels
    selected : list
        List with the names of the selected parameters to plot. Names match the names in the dataFrame. TO use all parameters, set selected to list(pars.keys())
    x_key : string
        Key in the dataframe for the 'x' axis data
    xlabel : string
        Label for the x-axis. Format: parameter [unit]
    ylabel : string
        Label for the y-axis. Format: parameter [unit]
    xscale : string
        Scale of the x-axis. E.g linear or log
    yscale : string
        Scale of the y-axis. E.g linear or log
    title : string
        Title of the plot
    ax : axes
        Axes object for the plot
    plot_funcs : Any
        Type of plot, e.g. standard plot or scatter
    x_error : List, optional
        List with error on the x parameter, by default []
    y_error : List, optional
        List with the error on the y parameter, by default []
    legend : bool, optional
        Toggle between showing the legend in the plot, by default True
    error_fmt : str, optional
        Format of the errorbars, by default '-'
    Returns
    -------
    axes
        Updated Axes object for the plot
    """
    i = 0
    for y_var in selected: # Plot the curve for every parameter specified in pars
        if (sum(data[y_var]) != 0):
            if plot_funcs == plt.errorbar:
                if len(x_error) == 0 and len(y_error) != 0:
                    plot_funcs(data[x_key], data[y_var], label=pars[y_var], yerr = y_error, color=plot_def.color[i], fmt=error_fmt)
                elif len(x_error) != 0 and len(y_error) != 0:
                    plot_funcs(data[x_key], data[y_var], label=pars[y_var], xerr = x_error, color=plot_def.color[i], fmt=error_fmt)
                elif len(x_error) != 0 and len(y_error) != 0:
                    plot_funcs(data[x_key], data[y_var], label=pars[y_var], xerr = x_error, yerr = y_error, color=plot_def.color[i], fmt=error_fmt)
                else:
                    plot_funcs(data[x_key], data[y_var], label=pars[y_var], color=plot_def.color[i], fmt=error_fmt)
            else: 
                 plot_funcs(data[x_key], data[y_var], label=pars[y_var], color=plot_def.color[i])               
        i += 1
    # Set figure and axis properties
    if legend:
        ax.legend()
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_xscale(xscale)
    ax.set_yscale(yscale)
    ax.set_title(title)
    return ax

def plot_result_colorbar_single(x,y,weight, ax,fig, xlabel, ylabel, weight_label, weight_norm, title, xscale='linear', yscale='linear'):
    """Show a x,y plot but use a colorbar to indicate a third parameter, similar to adding a weight to each point.

    Parameters
    ----------
    x : array
        Array with the x-data to plot
    y : array
        Array with the y data to plot
    weight : array
        Array with the weights for the colorbar
    ax : axes
        Axes object for the plot
    fig : figure
        Figure object for the plot
    xlabel : string
        label for the x-axis
    ylabel : string
        label for the y-axis
    weight_label : string
        label for the weights/colorbar
    weight_norm : string
        the scale for the colorbar, 'linear' or 'log'
    title : string
        Title of the plot
    xscale : str, optional
        set the scale of the x-axis, by default 'linear'
    yscale : str, optional
        set the scale of the y-axis, by default 'linear'

    Returns
    -------
    ax,fig
        return the axes,figure objects
    """

    # Divide the arrays into segments to color each segment individually
    points = np.array([x, y]).T.reshape(-1, 1, 2)
    segments = np.concatenate([points[:-1], points[1:]], axis=1)

    # Normalize the array for the colorbar (the weights), use either a log or lin scale
    if weight_norm == 'log':
        norm = matplotlib.colors.LogNorm(weight.min(), weight.max())
    else:
        norm = matplotlib.colors.Normalize(weight.min(), weight.max())

    lc = LineCollection(segments, cmap='plasma', norm=norm)
    lc.set_array(weight)

    # Create a colorbar plot
    line = ax.add_collection(lc)
    fig.colorbar(line, ax=ax, label=weight_label)          

    ax.margins(0.04) # Set some padding around the curve in the figure, because the method used here, puts the edges of the curve on the axis or takes the normalized axes.

    # Set figure properties
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_xscale(xscale)
    ax.set_yscale(yscale)
    ax.set_title(title)
    
    return ax, fig
    

def plot_result_twinx(data, pars, selected_1, selected_2, x_key, xlabel, ylabel_1, ylabel_2, xscale, yscale_1, yscale_2, title,ax1,ax2, plot_funcs, 
                        x_error=[], y_error_1 = [], y_error_2 = []):
    """Plot data on two y axis with a shared x axis. DIfferentiate between regular and errorbar plots.

    Parameters
    ----------
    data : DataFrame
        Data to plot
    pars : dict
        Dict with all potential parameters to plot. Keys represent the names in the dataFrama, values are the corresponding labels
    selected_1 : List
        List with the names of the selected parameters to plot on the left y axis.
        Names match the names in the dataFrame. TO use all parameters, set selected to list(pars.keys())
    selected_2 : List
        List with the names of the selected parameters to plot on the right y axis.
        Names match the names in the dataFrame. TO use all parameters, set selected to list(pars.keys())    x_key : _type_
        _description_
    xlabel : string
        Label for the x-axis. Format: parameter [unit]
    ylabel_1 : string
        Label for the left y-axis. Format: parameter [unit]
    ylabel_2 : _type_
        Label for the right y-axis. Format: parameter [unit]
    xscale : string
        Scale of the x-axis. E.g linear or log
    yscale_1 : string
        Scale of the left y-axis. E.g linear or log
    yscale_2 : string
        Scale of the right y-axis. E.g linear or log
    title : string
        Title of the plot
    ax1 : axes
        Axes object for the plot
    ax2 : axes
        Axes object for the plot (right y axis)
    plot_funcs : any
        Type of plot, e.g. standard plot or scatter
    x_error : list, optional
        _description_, by default []
    y_error_1 : list, optional
        List with error on the x parameter, by default []
    y_error_2 : list, optional
        List with error on the x parameter, by default []
    """
    
    # Set the x-axis
    ax1.set_xlabel(xlabel)
    ax1.set_xscale(xscale)

    ax1.set_title(title)

    # Set the left y axis
    ax1.set_ylabel(ylabel_1)
    ax1.set_yscale(yscale_1)

    #LEFT axis
    i = 0
    for y_var in selected_1:
        if (sum(data[y_var]) != 0):
            if plot_funcs == plt.errorbar:
                # Errorbar plot
                ax1.errorbar(data[x_key], data[y_var], yerr=y_error_1, label=pars[y_var],color=plot_def.color[i])
            else:
                if plot_funcs == plt.plot:
                    # Line plot
                    ax1.plot(data[x_key], data[y_var], label=pars[y_var],color=plot_def.color[i])
                elif plot_funcs == plt.scatter:
                    # Scatter plot
                    ax1.scatter(data[x_key], data[y_var], label=pars[y_var],color=plot_def.color[i])
        i += 1
    
    # RIGHT axis

    # Set the right y-axis
    ax2.set_ylabel(ylabel_2,color=plot_def.color[3])
    ax2.tick_params(axis='y', labelcolor=plot_def.color[3])
    ax2.set_yscale(yscale_2)

    i = 0
    for y_var in selected_2:
        if (sum(data[y_var]) != 0):
            if plot_funcs == plt.errorbar:
                # Errorbar plot
                ax2.errorbar(data[x_key], data[y_var], yerr = y_error_2, label=pars[y_var],color=plot_def.color[3],linestyle='dashed')
            else:
                if plot_funcs == plt.plot:
                    # Line plot
                    ax2.plot(data[x_key], data[y_var], label=pars[y_var],color=plot_def.color[3],linestyle='dashed')
                elif plot_funcs == plt.scatter:
                    # Scatter plot
                    ax2.scatter(data[x_key], data[y_var], label=pars[y_var],color=plot_def.color[3],linestyle='dashed')
    i += 1

    # Find the plotted lines and their labels on both axes
    lines, labels = ax1.get_legend_handles_labels()
    lines2, labels2 = ax2.get_legend_handles_labels()
    ax2.legend(lines + lines2, labels + labels2, loc=0)
    return ax1



""" Test the IMPS module of pySIMsalabim """

######### Package Imports #########################################################################
import os, sys, uuid 
import pandas as pd
import matplotlib.pyplot as plt
from joblib import Parallel, delayed
try :
    import pySIMsalabim as sim
except ImportError:
    # Add the parent directory to the system path
    sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '../..')))
    import pySIMsalabim as sim
from pySIMsalabim.experiments.imps import *

######### Test Functions #########################################################################

def test_run_IMPS_simu():
    """ Test the run_IMPS_simu function """
    # Set the path to the simulation setup file
    if os.path.exists('SIMsalabim'):
        cwd = os.path.abspath('.')
    else:
        cwd = os.path.abspath('../..')
    zimt_device_parameters = os.path.join(cwd, 'SIMsalabim','ZimT','simulation_setup.txt')
    session_path = os.path.join(cwd, 'SIMsalabim','ZimT')
    # Set the IMPS parameters
    tVG_name = os.path.join(session_path,'tVG.txt')
    f_min = 1e2
    f_max = 1e6
    f_steps = 20
    V = 0
    G_frac = 1
    GStep = 0.05
    # Run the IMPS simulation
    ret, mess = run_IMPS_simu(zimt_device_parameters, session_path, f_min, f_max, f_steps, V, G_frac, GStep, run_mode=False, tVG_name =tVG_name, output_file = 'freqY.dat', tj_name = 'tj.dat')
    # Clean up the output
    sim.clean_all_output(session_path)
    sim.clean_up_output('freqY',session_path)
    sim.delete_folders('tmp',session_path)
    # Check the output
    assert ret == 0, 'IMPS simulation failed'

def test_IMPS_parallel():
    """ Test the IMPS_parallel function """
    try:
        if os.path.exists('SIMsalabim'):
            cwd = os.path.abspath('.')
        else:
            cwd = os.path.abspath('../..')
        zimt_device_parameters = os.path.join(cwd, 'SIMsalabim','ZimT','simulation_setup.txt')
        session_path = os.path.join(cwd, 'SIMsalabim','ZimT')

        def run(G_frac,ID): # function to run the IMPS simulation in parallel
            cmd_pars = [{'par': 'R_series', 'val': '1e-4'},{'par': 'R_shunt', 'val': '1e2'},{'par': 'minAcc', 'val': '1e-1'},{'par': 'maxAcc', 'val': '1e-1'}]
            tVG_name = os.path.join(session_path,'tVG.txt')
            f_min = 1e2
            f_max = 1e6
            f_steps = 20
            V = 0
            GStep = 0.05
            print('Running')
            ret, mess = run_IMPS_simu(zimt_device_parameters, session_path, f_min, f_max, f_steps, V, G_frac, GStep, run_mode=False, tVG_name =tVG_name, output_file = 'freqY.dat', tj_name = 'tj.dat', UUID = ID,cmd_pars=cmd_pars,threadsafe=True)

        Gfracs = [ 0.1,0.5,1]
        ID_list = [str(uuid.uuid4()) for i in range(len(Gfracs))]
        Parallel(n_jobs=min(len(Gfracs),10))(delayed(run)(G_frac,ID) for G_frac,ID in zip(Gfracs,ID_list))

        # print('Plotting')
        # plt.figure()
        # for ID,G in zip(ID_list,Gfracs):
        #     data_tj = pd.read_csv(os.path.join(session_path,'freqY_'+str(ID)+'.dat'), sep=r'\s+')
        #     # data_tj = pd.read_csv(os.path.join(session_path,'freqZ.dat'), sep=r'\s+')
        #     # pars = {'Jext' : str(speed)} #'$J_{ext}$'}
        #     plt.plot(data_tj['V'],1/(data_tj['C']*1e2)**2,label = str(G),marker='o')
        #     plt.legend()
        # plt.show()

        sim.clean_all_output(session_path)
        sim.clean_up_output('freqY',session_path)
        sim.clean_up_output('Gfracs',session_path)
        sim.delete_folders('tmp',session_path)
        assert True
    except Exception as e:
        print(e)
        assert False, 'IMPS parallel simulation failed'


if __name__ == '__main__':
    test_run_IMPS_simu()
    test_IMPS_parallel()
    print('All IMPS tests passed')


import sys, os, subprocess, shutil, json, requests, zipfile, io
from packaging import version
from pathlib import Path
from getpass import getpass

######### Parameter Definitions ###################################################################
cwd = Path.cwd()
folder_name = 'kostergroup-SIMsalabim-'
min_fpc_version = '3.2.0'

######### Function Definitions ####################################################################
def cmd_yes_no_question(question, default = "yes"):
    """_summary_

    Parameters
    ----------
    question : str
        question to ask the user
    default : str, optional
        default answer, by default "yes"

    Returns
    -------
    bool
        whether the default answer is valid or not

    Raises
    ------
    ValueError
        If the default answer is not valid
    """    
    # valid answers (yes/no)
    valid = {'yes' : True, 'y': True, 'ye': True, 'no': False, 'n': False}
    # Set the default answer.
    if default is None:
        prompt = " [y/n] "
    elif default == 'yes':
        prompt = " [Y/n] "
    elif default == 'no':
        prompt = " [y/N] "
    else:
        raise ValueError(f"Invalid default answer: {default}\n")
    
    while True:
        # Capture user input
        sys.stdout.write(question + prompt)
        choice = input()
        # convert the input to lowercase
        choice = choice.lower()
        if default is not None and choice == "":
            # Use default value
            return valid[default]
        elif choice in valid:
            # Use user input
            return valid[choice]
        else:
            # Incorrect input
            sys.stdout.write('Please respond with "y" or "n"\n')

def clear_current_working_directory(cwd, folder_name='kostergroup-SIMsalabim-'):
    """_summary_

    Parameters
    ----------
    cwd : string
        Current working directory
    folder_name : string, optional
        Name of the folder to download, by default 'kostergroup-SIMsalabim-'

    Returns
    -------
    bool
        Whether the folder was removed or not

    """    
    # Clear the current working directory
    for dirpath, dirnames, files in os.walk(cwd):
        for dirname in dirnames:
            if dirname.startswith(folder_name):
                result = cmd_yes_no_question(f"Are you sure you want to overwrite the {dirname} folder?")
                if result == True:
                    shutil.rmtree(os.path.join(cwd,dirname))
                    print(f"Found and removed a folder named {dirname}")
                    return True
                elif result is False:
                    print(f"Not allowed to write into folder {dirname}")
                    return False
            else:
                # print(f"No folder found named SIMsalabim, continue.")
                return True

def get_SIMsalabim_source(cwd, folder_name='kostergroup-SIMsalabim-',verbose=False):
    """ Get the latest release from the Kostergroup Github

    Parameters
    ----------
    cwd : string
        Current working directory
    folder_name : string, optional
        Name of the folder to download, by default 'kostergroup-SIMsalabim-'
    verbose : bool, optional
        Print verbose output, by default False

    Returns
    -------
    int

    0 : Success
    2 : Failed
    3 : Failed

    """    
    if verbose:
        print("Getting the latest release from the Kostergroup Github")
    # Get the SIMsalabim source code.
    if os.path.exists(os.path.join(cwd, 'SIMsalabim')):
    # Pop out dialog box to confirm overwriting
        result = cmd_yes_no_question("Are you sure you want to overwrite the 'SIMsalabim' folder?")
        if result == True:
            # Remove folder
            shutil.rmtree(os.path.join(cwd, 'SIMsalabim'))

            # # Get the files from the latest release
            url = 'https://api.github.com/repos/kostergroup/SIMsalabim/zipball'
            response = requests.get(url)

            # Open the zip file
            z = zipfile.ZipFile(io.BytesIO(response.content))

            # Extract all the files
            z.extractall(path=cwd)

            for dirpath, dirnames, files in os.walk(cwd):
                for dirname in dirnames:
                    if dirname.startswith(folder_name):
                        # Rename folder
                        shutil.move(os.path.join(cwd, dirname), os.path.join(cwd, 'SIMsalabim'))
                        print("\nGot the latest release of SIMsalabim")
                        return 0
        else:
            print('We keep the current SIMsalabim version')
            return 0
    else:
        # # Get the files from the latest release
        url = 'https://api.github.com/repos/kostergroup/SIMsalabim/zipball'
        response = requests.get(url)

        # Open the zip file
        z = zipfile.ZipFile(io.BytesIO(response.content))

        # Extract all the files
        z.extractall(path=cwd)

        for dirpath, dirnames, files in os.walk(cwd):
            for dirname in dirnames:
                if dirname.startswith(folder_name):
                    # print(f"Found a folder named {dirname}")
                    # Rename folder
                    shutil.move(os.path.join(cwd, dirname), os.path.join(cwd, 'SIMsalabim'))
                    print("\nGot the latest release of SIMsalabim")
                    return 0

def get_SIMsalabim_assets(cwd,folder_name='kostergroup-SIMsalabim-',verbose=False):
    """ Get the latest compiled binaries from the Kostergroup Github

    Parameters
    ----------
    cwd : string
        Current working directory
    folder_name : string, optional
        Name of the folder to download, by default 'kostergroup-SIMsalabim-'
    verbose : bool, optional
        Print verbose output, by default False

    Returns
    -------
    int

    0 : Success
    1 : Success
    2 : Failed
    3 : Failed

    """    
    
    if verbose:
        print("Getting the latest compiled binaries from the Kostergroup Github")
    # Get the assets from the latest release
    url = "https://api.github.com/repos/kostergroup/SIMsalabim/releases/latest"
    response = requests.get(url)
    data = json.loads(response.text)

    for asset in data["assets"]:
        download_url = asset["browser_download_url"]
        filename = asset["name"]
        response = requests.get(download_url)
        open(os.path.join(cwd,filename), "wb").write(response.content)

    for dirpath, dirnames, files in os.walk(cwd):

        for filename in files:
            if filename.startswith('simss') and os.path.exists(os.path.join(cwd, filename)):
                print(f"Found a folder named {filename}")
                # Rename folder
                shutil.move(os.path.join(cwd, filename), os.path.join(cwd, 'SIMsalabim','SimSS',filename))
            elif filename.startswith('zimt') and os.path.exists(os.path.join(cwd, filename)):
                print(f"Found a folder named {filename}")
                # Rename folder
                shutil.move(os.path.join(cwd, filename), os.path.join(cwd, 'SIMsalabim','ZimT',filename))
            else:
                pass
    return 1

def use_SIMsalabim_source(cwd, folder_name='kostergroup-SIMsalabim-'):
    """ Use the SIMsalabim source code

    Parameters
    ----------
    cwd : string
        Current working directory
    folder_name : string, optional
        Name of the folder to download, by default 'kostergroup-SIMsalabim-'

    Returns
    -------
    int

    0 : Success
    2 : Failed
    3 : Failed

    """    
    # Clear the working directory. TEMP disabled
    # result = clear_current_working_directory(cwd, folder_name)
    result = True
    if result == True:
        result_get = get_SIMsalabim_source(cwd, folder_name)
        return result_get
    elif result == False:
        print('Script terminated manually')
        return 3
    else: 
        print('Failed')
        return 2

def install_SIMsalabim(cwd, folder_name='kostergroup-SIMsalabim-',verbose=False):
    """Install SIMsalabim in the current working directory

    Parameters
    ----------
    cwd : string
        Current working directory
    folder_name : string, optional
        Name of the folder to download, by default 'kostergroup-SIMsalabim-'
    verbose : bool, optional
        Print verbose output, by default False

    Returns
    -------
    int

    0 : Success
    1 : Success
    2 : Failed
    3 : Failed

    """    

    # Check if fpc installed
    if verbose:
        print("Checking for Free Pascal Compiler (fpc) package")
    if shutil.which("fpc") is not None:
        # fpc is installed, check the version and print to stdout.
        result = subprocess.run(["fpc", "-iV"], stdout=subprocess.PIPE, text=True)
        fpc_version = result.stdout

        # remove possible newline character
        if '\n' in fpc_version:
            fpc_version = fpc_version.strip('\n')

        # fpc version must be larger than min_fpc_version
        if version.parse(fpc_version) > version.parse(min_fpc_version):
            if verbose:
                print(f'Free Pascal Compiler (fpc) is installed with version >= {min_fpc_version}\n')
            result_fpc = use_SIMsalabim_source(cwd, folder_name)
            # sys.exit(result_fpc)

            # Compile the programs
            # compile simss with fpc in os.path.join(path2prog,'SimSS')
            os.chdir(os.path.join(cwd,'SIMsalabim','SimSS'))
            os.system('fpc simss.pas')
            os.chdir(cwd)

            # compile zimt with fpc in os.path.join(path2prog,'ZimT')
            cwd = os.getcwd() # Get current working directory
            os.chdir(os.path.join(cwd,'SIMsalabim','ZimT'))
            os.system('fpc zimt.pas')
            os.chdir(cwd)

            if verbose:
                print('SIMsalabim programs have been compiled successfully!')

        else: 
            # fpc version requirement not met
            print(f'Installed Free Pascal Compiler (fpc) version is {fpc_version}, but must be at least 3.2.0\n')
            result = cmd_yes_no_question("Do you want to continue with the pre-compiled binaries (y) or abort and update the Free Pascal Compiler (n) (recommended)", 'no')
            if result is True:
                # download assets
                print("\ndownload assets")
                result_assets = get_SIMsalabim_assets(cwd, verbose)
                sys.exit(result_assets)
            elif result is False:
                # return and exit
                sys.exit(3)
    else:
        # fpc is not installed.
        print("Free Pascal Compiler is not installed.\n")
        result = cmd_yes_no_question("Do you want to continue with the pre-compiled binaries (y) or abort and install the Free Pascal Compiler (n) (recommended)", 'no')
        if result is True:
            # download assets
            print("\ndownload assets")
            result_assets = get_SIMsalabim_assets(cwd, verbose)
            sys.exit(result_assets)
        elif result is False:
            # return and exit
            sys.exit(3)

def install_fpc_Linux():
    """ Install Free Pascal Compiler on Linux
    """   
    # check if system is Linux
    if os.name != 'posix':
        print('This function is only available on Linux systems')
        sys.exit()
    
    # check if apt-get is available
    if shutil.which("apt-get") is None:
        print('apt-get is not available on this system')
        sys.exit()

    # Install Free Pascal Compiler
    password = getpass('Sudo password:')
    # run the command sudo apt-get install fp-compiler and enter the password and print the output
    os.system('echo {} | sudo -S apt-get install fp-compiler'.format(password))

# def install_parallel_Linux():
#     """ Install GNU Parallel on Linux
#     """   
#     # check if system is Linux
#     if os.name != 'posix':
#         print('This function is only available on Linux systems')
#         sys.exit()
    
#     # check if apt-get is available
#     if shutil.which("apt-get") is None:
#         print('apt-get is not available on this system')
#         sys.exit()

#     # Install GNU Parallel
#     password = getpass('Sudo password:')
#     # run the command sudo apt-get install parallel and enter the password and print the output
#     os.system('echo {} | sudo -S apt-get install parallel'.format(password))

# def fpc_prog(cwd,prog_name):
#     """_summary_

#     Parameters
#     ----------
#     cwd : string
#         Current working directory
#     prog_name : string
#         Name of the program to compile

#     """
#     # Check if fpc installed
#     if verbose:
#         print("Checking for Free Pascal Compiler (fpc) package")
#     if shutil.which("fpc") is not None:
#         # fpc is installed, check the version and print to stdout.
#         result = subprocess.run(["fpc", "-iV"], stdout=subprocess.PIPE, text=True)
#         fpc_version = result.stdout

#         # remove possible newline character
#         if '\n' in fpc_version:
#             fpc_version = fpc_version.strip('\n')

#         # fpc version must be larger than min_fpc_version
#         if version.parse(fpc_version) > version.parse(min_fpc_version):
#             if verbose:
#                 print(f'Free Pascal Compiler (fpc) is installed with version >= {min_fpc_version}\n')
#             # Compile the programs
#             os.chdir(os.path.join(cwd,prog_name))
#             os.system(f'fpc {prog_name}.pas')
#             os.chdir(cwd)
#         else:
#             # fpc version requirement not met
#             print(f'Installed Free Pascal Compiler (fpc) version is {fpc_version}, but must be at least 3.2.0\n')
#             result = cmd_yes_no_question("Do you want to continue with the pre-compiled binaries (y) or abort and update the Free Pascal Compiler (n) (recommended)", 'no')
#             if result is True:
#                 # download assets
#                 print("\ndownload assets")
#                 result_assets = get_SIMsalabim_assets(cwd, verbose)
#                 sys.exit(result_assets)
#             elif result is False:
#                 # return and exit
#                 sys.exit(3)
#     else:
#         # fpc is not installed.
#         print("Free Pascal Compiler is not installed.\n")
        
#         sys.exit(3)

def fpc_prog(cwd,prog_name,show_term_output=True,force_fpc=True,verbose=True):
    """Compile program using fpc

    Parameters
    ----------
    cwd : string
        String of the absolute path to the program
    prog_name : string
        Name of the program to compile
    show_term_output : bool, optional
        show terminal output from the compilation, by default True
    force_fpc : bool, optional  
        force recompile with fpc even if compiled program already exists, by default True
    verbose : bool, optional
        print output of the compilation, by default True
    """   

    is_windows = (os.name == 'nt')          # Check if we are on Windows
    cwd = str(cwd)                  # Convert to string

    # check if
    compiled = False
    while not compiled:
        if shutil.which("fpc") is not None:
            # fpc is installed, check the version and print to stdout.
            result = subprocess.run(["fpc", "-iV"], stdout=subprocess.PIPE, text=True)
            fpc_version = result.stdout

            # remove possible newline character
            if '\n' in fpc_version:
                fpc_version = fpc_version.strip('\n')

            # fpc version must be larger than min_fpc_version
            if version.parse(fpc_version) >= version.parse(min_fpc_version):
                if verbose:
                    print(f'Free Pascal Compiler (fpc) is installed with version >= {min_fpc_version}\n') 
                # Check if the program is already compiled
                if (os.path.isfile(os.path.join(cwd,prog_name+'.exe')) and is_windows) or (os.path.isfile(os.path.join(cwd,prog_name)) and not is_windows):
                    if force_fpc:
                        if show_term_output == True:
                            output_direct = None
                        else:
                            output_direct = subprocess.DEVNULL
                        try:
                            subprocess.check_call(['fpc', prog_name.lower()+'.pas'], encoding='utf8', stdout=output_direct, cwd=cwd, shell=is_windows)
                        except subprocess.CalledProcessError:
                            print('Error compiling '+prog_name+' in '+cwd)
                            raise ChildProcessError
                        if verbose:
                            print('\n'+prog_name+' already existed but was recompiled'+'\n')
                    else:
                        if verbose:  
                            print('\n'+prog_name+' already compiled')
                
                else: # Compile the program
                    if show_term_output == True:
                        output_direct = None
                    else:
                        output_direct = subprocess.DEVNULL
                    try:
                        subprocess.check_call(['fpc', prog_name.lower()+'.pas'], encoding='utf8', stdout=output_direct, cwd=cwd, shell=is_windows)
                        if verbose:
                            print('\n'+prog_name+' was not compiled so we did it!'+'\n')
                    except subprocess.CalledProcessError:
                        print('Error compiling '+prog_name+' in '+cwd)
                        raise ChildProcessError 
                compiled = True
            else:
                # fpc version requirement not met
                print(f'Installed Free Pascal Compiler (fpc) version is {fpc_version}, but must be at least 3.2.0\n')
                print('Please update the Free Pascal Compiler to the latest version')
                sys.exit(3)
        else:
            # fpc is not installed.
            print("Free Pascal Compiler is not installed.\n")
            result = cmd_yes_no_question("Do you want to continue and install the Free Pascal Compiler (only works on Linux) (y) or abort (n) (recommended)", 'no')
            if result is True:
                # install fpc
                install_fpc_Linux()
                # sys.exit(0)
            elif result is False:
                # return and exit
                sys.exit(3)

        


  
######### Script ##################################################################################
if __name__ == "__main__":

    install_SIMsalabim(cwd,verbose=True)
    sys.exit(0)


""" Test the CV module of pySIMsalabim """

######### Package Imports #########################################################################
import os, sys, uuid 
import pandas as pd
import matplotlib.pyplot as plt
from joblib import Parallel, delayed
try :
    import pySIMsalabim as sim
except ImportError:
    # Add the parent directory to the system path
    sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '../..')))
    import pySIMsalabim as sim
from pySIMsalabim.experiments.CV import *

######### Test Functions #########################################################################

def test_run_CV_simu():
    """ Test the run_CV_simu function """
    # Set the path to the simulation setup file
    if os.path.exists('SIMsalabim'):
        cwd = os.path.abspath('.')
    else:
        cwd = os.path.abspath('../..')
    zimt_device_parameters = os.path.join(cwd, 'SIMsalabim','ZimT','simulation_setup.txt')
    session_path = os.path.join(cwd, 'SIMsalabim','ZimT')
    # Set the CV parameters
    tVG_name = os.path.join(session_path,'tVG.txt')
    freq = 1e4
    V_min = -0.8
    V_max = 1.2
    V_step = 0.1
    del_V = 0.01
    G_frac = 0
    # Run the CV simulation
    ret, mess = run_CV_simu(zimt_device_parameters, session_path, freq, V_min, V_max, V_step ,G_frac, del_V,run_mode=False, tVG_name=tVG_name, output_file = 'CapVol.dat', tj_name = 'tj.dat')
    print(ret,mess)
    # Clean up the output
    sim.clean_all_output(session_path)
    sim.clean_up_output('CapVol',session_path)
    sim.delete_folders('tmp',session_path)
    # Check the output
    assert ret == 0, 'CV simulation failed'
    
def test_CV_parallel():
    """ Test the CV_parallel function """
    try:
        if os.path.exists('SIMsalabim'):
            cwd = os.path.abspath('.')
        else:
            cwd = os.path.abspath('../..')
        zimt_device_parameters = os.path.join(cwd, 'SIMsalabim','ZimT','simulation_setup.txt')
        session_path = os.path.join(cwd, 'SIMsalabim','ZimT')

        def run(G_frac,ID): # function to run the CV simulation in parallel
            cmd_pars = [{'par': 'R_series', 'val': '1e-4'},{'par': 'R_shunt', 'val': '1e2'},{'par': 'minAcc', 'val': '1e-1'},{'par': 'maxAcc', 'val': '1e-1'}]
            tVG_name = os.path.join(session_path,'tVG.txt')
            freq = 1e4
            V_min = -0.8
            V_max = 1.2
            V_step = 0.1
            del_V = 0.01
            print('Running')
            ret, mess = run_CV_simu(zimt_device_parameters, session_path, freq, V_min, V_max, V_step ,G_frac, del_V,run_mode=False, tVG_name=tVG_name, output_file = 'CapVol.dat', tj_name = 'tj.dat', UUID = ID,cmd_pars=cmd_pars,threadsafe=True)

        Gfracs = [ 0.1,0.5,1]
        ID_list = [str(uuid.uuid4()) for i in range(len(Gfracs))]

        # run the CV simulation in parallel
        Parallel(n_jobs=min(len(Gfracs),10))(delayed(run)(G_frac,ID) for G_frac,ID in zip(Gfracs,ID_list))

        # print('Plotting')
        # plt.figure()
        # for ID,G in zip(ID_list,Gfracs):
        #     plot_capacitance(session_path, output_file=os.path.join(session_path,os.path.join(session_path,'CapVol_'+str(ID)+'.dat')))
        # plt.show()

        sim.clean_all_output(session_path)
        sim.clean_up_output('CapVol',session_path)
        sim.clean_up_output('Gfracs',session_path)
        sim.delete_folders('tmp',session_path)
        assert True
    except Exception as e:
        print(e)
        assert False, 'CV parallel simulation failed'


if __name__ == "__main__":
    test_run_CV_simu()
    test_CV_parallel()
    print('All CV tests passed')


""" Test the impedance module of pySIMsalabim """

######### Package Imports #########################################################################
import os, sys, uuid 
import pandas as pd
import matplotlib.pyplot as plt
from joblib import Parallel, delayed
try :
    import pySIMsalabim as sim
except ImportError:
    # Add the parent directory to the system path
    sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '../..')))
    import pySIMsalabim as sim
from pySIMsalabim.experiments.impedance import *

######### Test Functions #########################################################################
def test_run_impedance_simu():
    """ Test the run_impedance_simu function """
    # Set the path to the simulation setup file
    if os.path.exists('SIMsalabim'):
        cwd = os.path.abspath('.')
    else:
        cwd = os.path.abspath('../..')
    zimt_device_parameters = os.path.join(cwd, 'SIMsalabim','ZimT','simulation_setup.txt')
    session_path = os.path.join(cwd, 'SIMsalabim','ZimT')
    # Set the impedance parameters
    tVG_name = os.path.join(session_path,'tVG.txt')
    f_min = 1e-1
    f_max = 1e6
    f_steps = 20
    V_0 = 0
    del_V = 0.01
    G_frac = 0
    # Run the impedance simulation
    ret, mess = run_impedance_simu(zimt_device_parameters, session_path,  f_min, f_max, f_steps, V_0, G_frac, del_V, run_mode=False, tVG_name = tVG_name, output_file = 'freqZ.dat', tj_name = 'tj.dat')
    # Clean up the output
    sim.clean_all_output(session_path)
    sim.clean_up_output('freqZ',session_path)
    sim.delete_folders('tmp',session_path)
    # Check the output
    assert ret == 0, 'Impedance simulation failed'

def test_impedance_parallel():
    """ Test the impedance_parallel function """
    try:
        if os.path.exists('SIMsalabim'):
            cwd = os.path.abspath('.')
        else:
            cwd = os.path.abspath('../..')
        zimt_device_parameters = os.path.join(cwd, 'SIMsalabim','ZimT','simulation_setup.txt')
        session_path = os.path.join(cwd, 'SIMsalabim','ZimT')

        def run(G_frac,ID): # function to run the impedance simulation in parallel
            cmd_pars = [{'par': 'R_series', 'val': '1e-4'}]
            tVG_name = os.path.join(session_path,'tVG.txt')
            f_min = 1e-1
            f_max = 1e6
            f_steps = 20
            V_0 = 0
            del_V = 0.01
            print('Running')
            ret = run_impedance_simu(zimt_device_parameters, session_path, f_min, f_max, f_steps,  V_0, G_frac, del_V, run_mode=False, tVG_name = tVG_name, output_file = 'freqZ.dat', tj_name = 'tj.dat', UUID = ID,cmd_pars=cmd_pars,threadsafe=True)

        Gfracs = [ 0, 1]
        ID_list = [str(uuid.uuid4()) for i in range(len(Gfracs))]

        Parallel(n_jobs=min(len(Gfracs),10))(delayed(run)(G_frac,ID) for G_frac,ID in zip(Gfracs,ID_list))

        # print('Plotting')
        # plt.figure()
        # for ID,G in zip(ID_list,Gfracs):
        #     data_tj = pd.read_csv(os.path.join(session_path,'freqZ_'+str(ID)+'.dat'), sep=r'\s+')
        #     plt.loglog(data_tj['freq'],data_tj['C']*1e5,label=str(G),marker='o')
        #     plt.legend()
        # plt.show()

        sim.clean_all_output(session_path)
        sim.clean_up_output('freqZ',session_path)
        sim.clean_up_output('Gfracs',session_path)
        sim.delete_folders('tmp',session_path)
    except Exception as e:
        print('Error:',e)
        raise e


if __name__ == '__main__':
    test_run_impedance_simu()
    test_impedance_parallel()
    print('All impedance tests passed')
